@@ -35,6 +35,7 @@ _db_dio_table = None
 _db_dio_foreign_key_type = None
 _db_dio_schema_type = None
 # ================================================== database_utils.drawio - types
+_db_mysql_manager_type = None
 _db_mysql_database_type = None
 _db_column_type = None
 _db_column_sql_data_type = None
@@ -45,6 +46,7 @@ _db_mysql_select_query_type = None
 _db_mysql_update_query_type = None
 _db_mysql_delete_query_type = None
 _db_table_type = None
+_db_relationship_type = None
 
 
 
@@ -63,6 +65,8 @@ if TYPE_CHECKING:
     import colemen_utilities.database_utils.MySQL.Table.Table as _table
     _db_table_type = _TypeVar('_db_table_type', bound=_table.Table)
 
+    import colemen_utilities.database_utils.MySQL.Relationship.Relationship as _rel
+    _db_relationship_type = _TypeVar('_db_relationship_type', bound=_rel.Relationship)
 
     import colemen_utilities.database_utils.MySQL.DeleteQuery as _deleteQuery
     _db_mysql_delete_query_type = _TypeVar('_db_mysql_delete_query_type', bound=_deleteQuery.DeleteQuery)
@@ -76,6 +80,9 @@ if TYPE_CHECKING:
     import colemen_utilities.database_utils.MySQL.InsertQuery as _insertQuery
     _db_mysql_insert_query_type = _TypeVar('_db_mysql_insert_query_type', bound=_insertQuery.InsertQuery)
 
+    import colemen_utilities.database_utils.MySQL.DatabaseManager as _mysql_dbm
+    _db_mysql_manager_type = _TypeVar('_db_mysql_manager_type', bound=_mysql_dbm.DatabaseManager)
+
 
 
 
