@@ -356,9 +356,17 @@ class Column:
                 if hasattr(self._column.form,k):
                     print(f"form prop match found {k}")
                     setattr(self._column.form,k,v)
+                
+                if "validation_" in k:
+                    k = k.replace("validation_","")
+                    if k == "email":
+                        k = "is_email"
+                    if k == "unix_timestamp":
+                        k = "is_timestamp"
                 if hasattr(self._column.validation,k):
                     setattr(self._column.validation,k,v)
 
+
             if isinstance(self._column.table.raw_fulltext_indexes,(list)):
                 for ft in self._column.table.raw_fulltext_indexes:
                     if ft['column_name'] == self.column_name:
@@ -552,6 +560,46 @@ class Column:
                 self.column_default = value
             return value
 
+        @default.setter
+        def default(self,value):
+            '''
+                Set the Column's default property
+
+                `default`:None
+
+
+                Meta
+                ----------
+                `@author`: Colemen Atwood
+                `@created`: 12-20-2022 14:02:01
+                `@memberOf`: Column
+                `@property`: default
+            '''
+            self.column_default = value
+
+        @property
+        def flask_url_placeholder(self):
+            '''
+                Get this Column's flask_url_placeholder
+                
+                <{type}:{column_name}>
+
+                `default`:None
+
+
+                Meta
+                ----------
+                `@author`: Colemen Atwood
+                `@created`: 12-21-2022 08:32:10
+                `@memberOf`: Column
+                `@property`: flask_url_placeholder
+            '''
+            name = self.column_name
+            if self.is_hash_id is True:
+                name = self._column.table.primary_column_name
+            value = f"<{self.py_data_type}:{name}>"
+            return value
+
         # @property
         # def is_hash_id(self):
         #     '''
