@@ -30,7 +30,7 @@ import os
 
 import sys
 
-from typing import Iterable, Union as _Union
+from typing import Iterable, Union
 
 
 import mysql.connector as _mysqlConnector
@@ -86,6 +86,9 @@ class MySQLDatabase:
     _dbm:_db_mysql_manager_type = None
     '''The database manager instance used to manage multiple schemas'''
 
+    _table_relationships = None
+    '''A list of all table relationships retrieved from the database'''
+
 
     # ------------------------------- SUB-ENTITIES ------------------------------- #
 
@@ -190,6 +193,11 @@ class MySQLDatabase:
     #         table.cache.save()
         # _cfu.writer.to_json(f"{self.cache_path}/db_summary.json",self.summary)
 
+
+    # ---------------------------------------------------------------------------- #
+    #                                 GENERAL DATA                                 #
+    # ---------------------------------------------------------------------------- #
+
     @property
     def name(self):
         '''
@@ -208,45 +216,6 @@ class MySQLDatabase:
         return self.database
 
 
-    def register(self,entity):
-        '''
-            Used INTERNALLY to register table, column & relationship entities with this database.
-
-            ----------
-
-            Arguments
-            -------------------------
-            `entity` {any}
-                The entity to register with this database.
-
-            Meta
-            ----------
-            `author`: Colemen Atwood
-            `created`: 12-16-2022 10:01:12
-            `memberOf`: DatabaseManager
-            `version`: 1.0
-            `method_name`: register
-            * @TODO []: documentation for register
-        '''
-
-        if isinstance(entity,Table.Table):
-            table:_db_table_type = entity
-            self._tables[table.name] = table
-
-        if isinstance(entity,Relationship.Relationship):
-            rel:_db_relationship_type = entity
-            self._relationships.append(rel)
-
-        if isinstance(entity,Column):
-            col:_db_column_type = entity
-            self._columns.append(col)
-
-        # @Mstep [IF] if there is a manager associated to this database
-        if self._dbm is not None:
-            # @Mstep [] register there entity with the manager as well.
-            self._dbm.register(entity)
-
-
 
     # ---------------------------------------------------------------------------- #
     #                                  CONNECTION                                  #
@@ -395,7 +364,7 @@ class MySQLDatabase:
     # ---------------------------------------------------------------------------- #
 
 
-    def run(self, sql:str, args:_Union[list,dict]=False):
+    def run(self, sql:str, args:Union[list,dict]=False):
         '''
             Executes a query on the database.
 
@@ -565,6 +534,16 @@ class MySQLDatabase:
                 self.connect()
                 return self.execute_single_statement(sql,args,True)
 
+        except _mysqlConnector.connector.errors.OperationalError:
+            if isTimeoutRetry is True:
+                _log(f"{_traceback.format_exc()}","error")
+                _log(f"SQL: {sql}","error")
+            if isTimeoutRetry is False:
+                # _log(f"CONNECTION TIMED OUT")
+                self.cur = None
+                self.con = None
+                self.connect()
+                return self.execute_single_statement(sql,args,True)
         except _mysqlConnector.errors.DatabaseError:
             # print(f"ERROR: {err}", PRESET="FATAL_ERROR_INVERT")
             _log(f"{_traceback.format_exc()}","error")
@@ -729,28 +708,28 @@ class MySQLDatabase:
     # ---------------------------------------------------------------------------- #
 
 
-    def insert_query(self,table:_Union[str,_db_table_type])->_db_mysql_insert_query_type:
+    def insert_query(self,table:Union[str,_db_table_type])->_db_mysql_insert_query_type:
         '''Create a new insert query instance for the table provided.'''
         if isinstance(table,(str)):
             return InsertQuery(database=self,table_name=table)
         else:
             return InsertQuery(database=self,table=table)
 
-    def select_query(self,table:_Union[str,_db_table_type])->_db_mysql_select_query_type:
+    def select_query(self,table:Union[str,_db_table_type])->_db_mysql_select_query_type:
         '''Create a new select query instance'''
         if isinstance(table,(str)):
             return SelectQuery(database=self,table_name=table)
         else:
             return SelectQuery(database=self,table=table)
 
-    def update_query(self,table:_Union[str,_db_table_type])->_db_mysql_update_query_type:
+    def update_query(self,table:Union[str,_db_table_type])->_db_mysql_update_query_type:
         '''Create a new update query instance'''
         if isinstance(table,(str)):
             return UpdateQuery(database=self,table_name=table)
         else:
             return UpdateQuery(database=self,table=table)
 
-    def delete_query(self,table:_Union[str,_db_table_type])->_db_mysql_delete_query_type:
+    def delete_query(self,table:Union[str,_db_table_type])->_db_mysql_delete_query_type:
         '''Create a new delete query instance'''
         if isinstance(table,(str)):
             return DeleteQuery(database=self,table_name=table)
@@ -759,7 +738,6 @@ class MySQLDatabase:
 
 
 
-
     # ---------------------------------------------------------------------------- #
     #                                SUPPORT QUERIES                               #
     # ---------------------------------------------------------------------------- #
@@ -866,190 +844,9 @@ class MySQLDatabase:
         return result
 
 
-
-    def get_column_data(self,table_name:str,default=None)->Iterable[_db_column_type]:
-        '''
-            Retrieve the column meta data for all columns in the table provided.
-
-            ----------
-
-            Arguments
-            -------------------------
-            `table_name` {str}
-                The name of the table to retrieve column data about.
-
-            [`default`=None] {any}
-                The value to return if no columns are found.
-
-
-            Return {list}
-            ----------------------
-            A list of column instances if successful, the default otherwise.
-
-            Meta
-            ----------
-            `author`: Colemen Atwood
-            `created`: 12-09-2022 11:18:41
-            `memberOf`: MySQLDatabase
-            `version`: 1.0
-            `method_name`: get_column_data
-            * @xxx [12-09-2022 11:19:52]: documentation for get_column_data
-        '''
-        # if table_name in self._tables:
-        #     return self._tables[table_name]['columns']
-
-        if self.connect() is False:
-            return False
-        # @Mstep [] get the INFORMATION_SCHEMA.COLUMNS data from the database.
-        sql = "SELECT * FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA=$schema_name AND TABLE_NAME=$table_name"
-        args = {
-            "schema_name":self.database,
-            "table_name":table_name,
-        }
-        result = self.run_select(sql, args)
-
-        if isinstance(result,(list)) is False:
-            return default
-
-        keys = self.get_foreign_keys(table_name)
-
-
-        columns = []
-        for col in result:
-            for key in keys:
-                if col['COLUMN_NAME'] == key['COLUMN_NAME']:
-                    col['constraint'] = key
-            columns.append(col)
-
-
-        # import colemen_utilities.database_utils.MySQL.Column.Column as _col
-        output = columns
-        # output = []
-        # for col in columns:
-            # output.append(_col.Column(col))
-            # output.append(_col.Column(col))
-        if len(columns) == 0:
-            output = default
-        return output
-
-    def get_referential_constraint_by_name(self,name:str,table_name:str):
-        if self.connect() is False:
-            return False
-        sql = f'''
-            SELECT
-                *
-            FROM
-                INFORMATION_SCHEMA.referential_constraints
-            WHERE
-                TABLE_NAME = $table_name AND
-                CONSTRAINT_NAME = $name
-        '''
-        args = {
-            "table_name":table_name,
-            "name":name,
-        }
-        _log(f"MySQLDatabase.get_referential_constraint_by_name - {name}","magenta")
-        result = self.run_select(sql,args)
-        if isinstance(result,(list)) is False:
-            result = []
-        else:
-            if len(result) > 0:
-                result = result[0]
-        return result
-
-    def get_referential_constraint_by_table(self,table_name:str):
-        if self.connect() is False:
-            return False
-        sql = f'''
-            SELECT
-                *
-            FROM
-                INFORMATION_SCHEMA.referential_constraints
-            WHERE
-                TABLE_NAME = $table_name;'''
-        args = {
-            # "schema_name":self.database,
-            "table_name":table_name,
-        }
-        result = self.run_select(sql,args)
-        if isinstance(result,(list)) is False:
-            result = []
-
-        return result
-
-    def get_fulltext_indexes_by_table(self,table_name):
-        '''Retrieve all fulltext indexes on the table provided'''
-        sql = f'''
-            SELECT
-            `INDEX_NAME` as index_name,
-            `COLUMN_NAME` as column_name,
-            `NULLABLE` as nullable,
-            `INDEX_COMMENT` as index_comment,
-            `INDEX_TYPE` as index_type
-            FROM INFORMATION_SCHEMA.STATISTICS 
-            WHERE
-            `INDEX_TYPE`='FULLTEXT' AND
-            `table_name`=$table_name AND
-            `table_schema` = $schema_name;'''
-        args = {
-            "schema_name":self.database,
-            "table_name":table_name,
-        }
-        result = self.run_select(sql,args)
-        if isinstance(result,(list)) is False:
-            result = []
-        # self._tables[table_name]
-        # self._tables[table_name]['foreign_keys'] = result
-        return result
-
-    def get_foreign_keys(self,table_name:str):
-        '''
-            Retrieve a list of columns that are foreign keys in the table provided.
-            ----------
-
-            Arguments
-            -------------------------
-            `table_name` {str}
-                The name of the table to query.
-
-
-            Return {list}
-            ----------------------
-            A list of foreign key columns, the list is empty if there are None.
-
-            Meta
-            ----------
-            `author`: Colemen Atwood
-            `created`: 12-09-2022 11:24:42
-            `memberOf`: MySQLDatabase
-            `version`: 1.0
-            `method_name`: get_foreign_keys
-            * @xxx [12-09-2022 11:26:26]: documentation for get_foreign_keys
-        '''
-        # if table_name in self._tables:
-        #     return self._tables[[table_name]]['foreign_keys']
-        if self.connect() is False:
-            return False
-        sql = f'''
-            SELECT
-                TABLE_NAME,COLUMN_NAME,CONSTRAINT_NAME, REFERENCED_TABLE_NAME,REFERENCED_COLUMN_NAME,REFERENCED_TABLE_SCHEMA
-            FROM
-                INFORMATION_SCHEMA.KEY_COLUMN_USAGE
-            WHERE
-                TABLE_NAME = $table_name AND
-                TABLE_SCHEMA = $schema_name AND
-                REFERENCED_TABLE_NAME IS NOT NULL;
-        '''
-        args = {
-            "schema_name":self.database,
-            "table_name":table_name,
-        }
-        result = self.run_select(sql,args)
-        if isinstance(result,(list)) is False:
-            result = []
-        # self._tables[table_name]
-        # self._tables[table_name]['foreign_keys'] = result
-        return result
+    # ---------------------------------------------------------------------------- #
+    #                                TABLE UTILITIES                               #
+    # ---------------------------------------------------------------------------- #
 
     def get_table_data(self,table_name:str):
         '''
@@ -1324,12 +1121,346 @@ class MySQLDatabase:
 
 
 
+    # ---------------------------------------------------------------------------- #
+    #                               COLUMN UTILITIES                               #
+    # ---------------------------------------------------------------------------- #
+
+    def get_column_data(self,table_name:str,default=None)->Iterable[_db_column_type]:
+        '''
+            Retrieve the column meta data for all columns in the table provided.
+
+            ----------
+
+            Arguments
+            -------------------------
+            `table_name` {str}
+                The name of the table to retrieve column data about.
+
+            [`default`=None] {any}
+                The value to return if no columns are found.
+
+
+            Return {list}
+            ----------------------
+            A list of column instances if successful, the default otherwise.
+
+            Meta
+            ----------
+            `author`: Colemen Atwood
+            `created`: 12-09-2022 11:18:41
+            `memberOf`: MySQLDatabase
+            `version`: 1.0
+            `method_name`: get_column_data
+            * @xxx [12-09-2022 11:19:52]: documentation for get_column_data
+        '''
+        # if table_name in self._tables:
+        #     return self._tables[table_name]['columns']
+
+        if self.connect() is False:
+            return False
+        # @Mstep [] get the INFORMATION_SCHEMA.COLUMNS data from the database.
+        sql = "SELECT * FROM INFORMATION_SCHEMA.COLUMNS WHERE TABLE_SCHEMA=$schema_name AND TABLE_NAME=$table_name"
+        args = {
+            "schema_name":self.database,
+            "table_name":table_name,
+        }
+        result = self.run_select(sql, args)
+
+        if isinstance(result,(list)) is False:
+            return default
+
+        keys = self.get_foreign_keys(table_name)
+
+
+        columns = []
+        for col in result:
+            for key in keys:
+                if col['COLUMN_NAME'] == key['COLUMN_NAME']:
+                    col['constraint'] = key
+            columns.append(col)
+
+
+        # import colemen_utilities.database_utils.MySQL.Column.Column as _col
+        output = columns
+        # output = []
+        # for col in columns:
+            # output.append(_col.Column(col))
+            # output.append(_col.Column(col))
+        if len(columns) == 0:
+            output = default
+        return output
+
+    def get_referential_constraint_by_name(self,name:str,table_name:str):
+        if self.connect() is False:
+            return False
+        sql = f'''
+            SELECT
+                *
+            FROM
+                INFORMATION_SCHEMA.referential_constraints
+            WHERE
+                TABLE_NAME = $table_name AND
+                CONSTRAINT_NAME = $name
+        '''
+        args = {
+            "table_name":table_name,
+            "name":name,
+        }
+        _log(f"MySQLDatabase.get_referential_constraint_by_name - {name}","magenta")
+        result = self.run_select(sql,args)
+        if isinstance(result,(list)) is False:
+            result = []
+        else:
+            if len(result) > 0:
+                result = result[0]
+        return result
+
+    def get_referential_constraint_by_table(self,table_name:str):
+        if self.connect() is False:
+            return False
+        sql = f'''
+            SELECT
+                *
+            FROM
+                INFORMATION_SCHEMA.referential_constraints
+            WHERE
+                TABLE_NAME = $table_name;'''
+        args = {
+            # "schema_name":self.database,
+            "table_name":table_name,
+        }
+        result = self.run_select(sql,args)
+        if isinstance(result,(list)) is False:
+            result = []
+
+        return result
+
+    def get_fulltext_indexes_by_table(self,table_name):
+        '''Retrieve all fulltext indexes on the table provided'''
+        sql = f'''
+            SELECT
+            `INDEX_NAME` as index_name,
+            `COLUMN_NAME` as column_name,
+            `NULLABLE` as nullable,
+            `INDEX_COMMENT` as index_comment,
+            `INDEX_TYPE` as index_type
+            FROM INFORMATION_SCHEMA.STATISTICS 
+            WHERE
+            `INDEX_TYPE`='FULLTEXT' AND
+            `table_name`=$table_name AND
+            `table_schema` = $schema_name;'''
+        args = {
+            "schema_name":self.database,
+            "table_name":table_name,
+        }
+        result = self.run_select(sql,args)
+        if isinstance(result,(list)) is False:
+            result = []
+        # self._tables[table_name]
+        # self._tables[table_name]['foreign_keys'] = result
+        return result
+
+    def get_foreign_keys(self,table_name:str):
+        '''
+            Retrieve a list of columns that are foreign keys in the table provided.
+            ----------
+
+            Arguments
+            -------------------------
+            `table_name` {str}
+                The name of the table to query.
+
+
+            Return {list}
+            ----------------------
+            A list of foreign key columns, the list is empty if there are None.
+
+            Meta
+            ----------
+            `author`: Colemen Atwood
+            `created`: 12-09-2022 11:24:42
+            `memberOf`: MySQLDatabase
+            `version`: 1.0
+            `method_name`: get_foreign_keys
+            * @xxx [12-09-2022 11:26:26]: documentation for get_foreign_keys
+        '''
+        # if table_name in self._tables:
+        #     return self._tables[[table_name]]['foreign_keys']
+        if self.connect() is False:
+            return False
+        sql = f'''
+            SELECT
+                TABLE_NAME,COLUMN_NAME,CONSTRAINT_NAME, REFERENCED_TABLE_NAME,REFERENCED_COLUMN_NAME,REFERENCED_TABLE_SCHEMA
+            FROM
+                INFORMATION_SCHEMA.KEY_COLUMN_USAGE
+            WHERE
+                TABLE_NAME = $table_name AND
+                TABLE_SCHEMA = $schema_name AND
+                REFERENCED_TABLE_NAME IS NOT NULL;
+        '''
+        args = {
+            "schema_name":self.database,
+            "table_name":table_name,
+        }
+        result = self.run_select(sql,args)
+        if isinstance(result,(list)) is False:
+            result = []
+        # self._tables[table_name]
+        # self._tables[table_name]['foreign_keys'] = result
+        return result
+
+
+
+    # ---------------------------------------------------------------------------- #
+    #                                 RELATIONSHIPS                                #
+    # ---------------------------------------------------------------------------- #
+
+
+    @property
+    def get_table_relationships(self):
+        '''
+            Query the database to retrieve all relationships between tables in this schema specifically
+
+            Relationships from other schemas ares filtered out.
+            This will include relationships to a different schema.
+
+            `default`:None
+
+
+            Meta
+            ----------
+            `@author`: Colemen Atwood
+            `@created`: 12-12-2022 09:21:04
+            `@memberOf`: DatabaseManager
+            `@property`: summary
+        '''
+        if self._table_relationships is not None:
+            print(f"table relationships already populated")
+            return self._table_relationships
+        
+        if self._dbm is not None:
+            result = self._dbm.get_table_relationships
+        else:
+            sql = '''
+                SELECT
+                `REFERENCED_TABLE_NAME` as 'parent_table_name',                 -- Origin key table
+                `REFERENCED_TABLE_SCHEMA` as 'parent_table_schema',               -- Origin key schema
+                `REFERENCED_COLUMN_NAME` as 'parent_column_name',                 -- Origin key column
+                `TABLE_NAME` as 'child_table_name',                            -- Foreign key table
+                `TABLE_SCHEMA` as 'child_table_schema',                          -- Foreign key schema
+                `COLUMN_NAME` as 'child_column_name'
+                FROM
+                `INFORMATION_SCHEMA`.`KEY_COLUMN_USAGE`  -- Will fail if user don't have privilege
+                WHERE
+                `REFERENCED_TABLE_NAME` IS NOT NULL -- Only tables with foreign keys
+                ORDER BY REFERENCED_TABLE_NAME
+                ;'''
+            result = self._default_schema.run_select(sql)
+        print(f"result: {result}")
+        output = []
+        for r in result:
+            if r['parent_table_schema'] not in [self.name] and r['child_table_schema'] not in [self.name]:
+                continue
+            output.append(r)
+        self._table_relationships = output
+        return output
+
+
+    def get_table_children(self,table:Union[str,_db_table_type])->Iterable[_db_table_type]:
+        '''
+            Retrieve a list of tables that have a foreign key constraint referencing the table provided.
+
+            Essentially retrieving all tables that are children of the table provided
+
+            ----------
+
+            Arguments
+            -------------------------
+            `table` {str,Table}
+                The name of the parent table or the Table instance.
+
+
+            Return {list}
+            ----------------------
+            A list of tables associated as children to the table provided.
+
+            If None are found, the list is empty.
+
+            Meta
+            ----------
+            `author`: Colemen Atwood
+            `created`: 12-16-2022 11:56:16
+            `memberOf`: DatabaseManager
+            `version`: 1.0
+            `method_name`: get_table_children
+            * @xxx [12-16-2022 12:01:52]: documentation for get_table_children
+        '''
+        table_name = _table_to_name(table)
+        if self._dbm is not None:
+            return self._dbm.get_table_children(table_name)
+        else:
+            table = self.get_table(table_name)
+            if table is not None:
+                tables = []
+                rels = self.get_table_relationships
+                for rel in rels:
+                    if rel['parent_table_name'] == table_name:
+                        self.database.get_table(rel['child_table_schema'])
+                        if table is not None:
+                            tables.append(table)
+                return tables
+
+
+    def get_table_parents(self,table:Union[str,_db_table_type])->Iterable[_db_table_type]:
+        '''
+            Retrieve a list of tables that have a foreign key constraint referencing the table provided.
+
+            Essentially retrieving all tables that are children of the table provided
+
+            ----------
+
+            Arguments
+            -------------------------
+            `table` {str,Table}
+                The name of the parent table or the Table instance.
+
+
+            Return {list}
+            ----------------------
+            A list of tables associated as children to the table provided.
+
+            If None are found, the list is empty.
+
+            Meta
+            ----------
+            `author`: Colemen Atwood
+            `created`: 12-16-2022 11:56:16
+            `memberOf`: DatabaseManager
+            `version`: 1.0
+            `method_name`: get_table_children
+            * @xxx [12-16-2022 12:01:52]: documentation for get_table_children
+        '''
+        table_name = _table_to_name(table)
+        if self._dbm is not None:
+            return self._dbm.get_table_parents(table_name)
+        else:
+            table = self.get_table(table_name)
+            if table is not None:
+                tables = []
+                rels = self.get_table_relationships
+                for rel in rels:
+                    if rel['child_table_name'] == table_name:
+                        self.database.get_table(rel['parent_table_name'])
+                        if table is not None:
+                            tables.append(table)
+                return tables
+
     def get_all_relationships(self)->Iterable[_db_relationship_type]:
         table:_db_table_type
         for _,table in self._tables.items():
             for col in table.columns:
                 if col.data.is_foreign_key:
-                    _log("Column.relationship --- Relationship Found","magenta")
+                    # _log("Column.relationship --- Relationship Found","magenta")
                     value = Relationship.Relationship(
                         self,
                         col.table,
@@ -1512,6 +1643,10 @@ class MySQLDatabase:
         return output
 
 
+    # ---------------------------------------------------------------------------- #
+    #                                DATA VALIDATION                               #
+    # ---------------------------------------------------------------------------- #
+
     def correlate_to_table(self,table_name:str,data:dict,**kwargs):
         '''
             Correlate a dictionary of data to the columns of the table.
@@ -1683,6 +1818,52 @@ class MySQLDatabase:
 
 
 
+    # ---------------------------------------------------------------------------- #
+    #                                    SUPPORT                                   #
+    # ---------------------------------------------------------------------------- #
+
+    def register(self,entity):
+        '''
+            Used INTERNALLY to register table, column & relationship entities with this database.
+
+            ----------
+
+            Arguments
+            -------------------------
+            `entity` {any}
+                The entity to register with this database.
+
+            Meta
+            ----------
+            `author`: Colemen Atwood
+            `created`: 12-16-2022 10:01:12
+            `memberOf`: DatabaseManager
+            `version`: 1.0
+            `method_name`: register
+            * @TODO []: documentation for register
+        '''
+
+        if isinstance(entity,Table.Table):
+            table:_db_table_type = entity
+            self._tables[table.name] = table
+
+        if isinstance(entity,Relationship.Relationship):
+            rel:_db_relationship_type = entity
+            self._relationships.append(rel)
+
+        if isinstance(entity,Column):
+            col:_db_column_type = entity
+            self._columns.append(col)
+
+        # @Mstep [IF] if there is a manager associated to this database
+        if self._dbm is not None:
+            # @Mstep [] register there entity with the manager as well.
+            self._dbm.register(entity)
+
+
+
+
+
 
 def _to_statement_list(sql):
     sql = sql.replace(";", ";STATEMENT_END")
@@ -1873,4 +2054,11 @@ def new(**kwargs)->MySQLDatabase:
 
 
 
+def _table_to_name(table:Union[str,_db_table_type]):
+    '''returns a table name if the table instance is provided.'''
+    from colemen_utilities.database_utils.MySQL.Table import Table
+    if isinstance(table,Table.Table):
+        table = table.name
+    return table
+
 
